/*
    SPDX-FileCopyrightText: 2022 David Faure <faure@kde.org>
    SPDX-License-Identifier: LGPL-2.0-or-later
*/
#include "workerfactory.h"

namespace KIO
{

WorkerFactory::WorkerFactory(QObject *parent)
    : QObject(parent)
{
}

} // namespace KIO

#include "moc_workerfactory.cpp"
