/*
    This file is part of the KDE project
    SPDX-FileCopyrightText: 2015 Olivier Goffart <ogoffart@woboq.com>

    SPDX-License-Identifier: LGPL-2.0-only OR LGPL-3.0-only OR LicenseRef-KDE-Accepted-LGPL
*/

#include "koverlayiconplugin.h"

KOverlayIconPlugin::KOverlayIconPlugin(QObject *parent)
    : QObject(parent)
{
}

KOverlayIconPlugin::~KOverlayIconPlugin()
{
}

#include "moc_koverlayiconplugin.cpp"
